//**************************************************************************
//*                     This file is part of the                           *
//*                MMC - Mpxplay Multimedia Commander                      *
//*                   The source code of MMC is                            *
//*        (C) copyright 1998-2014 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: video playing by QT

//#define MPXPLAY_USE_DEBUGF 1
#define DISPQT_DEBUG_OUTPUT NULL //stdout
#define DISPQT_DEBUGOUT_EVENT NULL // stdout
#define DISPQT_DEBUGOUT_CURSOR stdout

#include <QtGui>
#include <QtWidgets>
#include "disp_qt.h"
#include "moc_video_qt.h"
#include "moc_mainwindow.h"

extern "C" {
 extern struct mainvars mvps;
 extern unsigned int playcontrol, refdisp, desktopmode, playstartmsec;
 extern keyconfig kb[];
}

DispQtVideoPlayer::DispQtVideoPlayer(MainWindow *main_window, QWidget *parent) : QWidget(parent)
{
	this->main_window = main_window;
	this->parent_widget = parent;
	this->is_wait_for_open = false;
	this->is_video_show = false;
	this->is_video_open = false;
	this->is_playing = false;
	this->is_wait_for_close = false;
	this->is_video_closed = false;
	this->qt_media_player = NULL;
	this->ptr_handle_curr = this->ptr_handle_new = this->ptr_handle_close = NULL;

	this->video_widget = new DispQtVideoWidget(main_window, this);
	this->video_mediaplayer_open();
	this->video_layout = new QHBoxLayout;
	this->video_layout->addWidget(this->video_widget);
	if(this->qt_media_player)
		this->qt_media_player->setVideoOutput(this->video_widget);
	this->video_layout->setMargin(1);
	this->video_layout->setSpacing(0);
	this->setLayout(this->video_layout);

	connect(this, SIGNAL(signal_video_open()), this, SLOT(video_open()));
	connect(this, SIGNAL(signal_video_close()), this, SLOT(video_close()));
	connect(this, SIGNAL(signal_video_play(bool)), this, SLOT(video_play(bool)));
	connect(this, SIGNAL(signal_video_pause()), this, SLOT(video_pause()));
	connect(this, SIGNAL(signal_video_switch_play()), this, SLOT(video_switch_play()));
	connect(this, SIGNAL(signal_video_seek_ms(qint64)), this, SLOT(video_seek_ms(qint64)));
	connect(this, SIGNAL(signal_video_set_fullscreen(bool)), this, SLOT(video_set_fullscreen(bool)));
	connect(this, SIGNAL(signal_video_switch_fullscreen()), this, SLOT(video_switch_fullscreen()));
	connect(this, SIGNAL(signal_video_apply_fullscreen(bool)), this, SLOT(video_apply_fullscreen(bool)));
	connect(this, SIGNAL(signal_video_check_windows()), this, SLOT(video_check_windows()));
	connect(this, SIGNAL(signal_video_apply_windows()), this, SLOT(video_apply_windows()));
	connect(this, SIGNAL(signal_video_event_stop()), this, SLOT(video_handle_event_stop()));
	connect(this, SIGNAL(signal_video_set_speed(int)), this, SLOT(video_set_speed(int)));
	connect(this, SIGNAL(signal_audio_set_volume(int)), this, SLOT(audio_set_volume(int)));
}

DispQtVideoPlayer::~DispQtVideoPlayer()
{
	if(this->is_video_open && this->qt_media_player){
		playstartmsec = (unsigned long)this->video_tell_ms();
		if(this->video_widget && this->video_widget->isFullScreen()) {
			this->main_window->video_fullscreen_switch(false);
			this->video_apply_fullscreen(false);
		}
		this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
		this->is_video_open = false;
		this->qt_media_player->stop();
		mvps.adone = ADONE_REOPEN;
		mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "close adone:%d ms:%d", mvps.adone, playstartmsec);
		this->mutex_qmedia.unlock();
	}
}

void DispQtVideoPlayer::closeEvent(QCloseEvent *e)
{
	if(this->is_video_open && this->qt_media_player){
		this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
		this->is_video_open = false;
		this->qt_media_player->stop();
		this->mutex_qmedia.unlock();
	}
	QWidget::closeEvent(e);
}

void DispQtVideoPlayer::video_mediaplayer_open(void)
{
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(this->qt_media_player){
		this->mutex_qmedia.unlock();
		return;
	}
	this->qt_media_player = new QMediaPlayer(this);
	if(this->video_widget)
		this->qt_media_player->setVideoOutput(this->video_widget);
	connect(this->qt_media_player, SIGNAL(mediaStatusChanged(QMediaPlayer::MediaStatus)), this, SLOT(status_changed(QMediaPlayer::MediaStatus)));
	connect(this->qt_media_player, SIGNAL(error(QMediaPlayer::Error)), this, SLOT(display_error_message(QMediaPlayer::Error)));
	this->mutex_qmedia.unlock();
}

void DispQtVideoPlayer::video_mediaplayer_close(void)
{
	this->video_cursor_select(QMediaPlayer::UnknownMediaStatus);
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(!this->qt_media_player) {
		this->mutex_qmedia.unlock();
		return;
	}
	this->is_playing = false;
	if(this->is_video_open)
		this->qt_media_player->stop();
	this->is_video_open = false;
	disconnect(this->qt_media_player, SIGNAL(mediaStatusChanged(QMediaPlayer::MediaStatus)), this, SLOT(status_changed(QMediaPlayer::MediaStatus)));
	disconnect(this->qt_media_player, SIGNAL(error(QMediaPlayer::Error)), this, SLOT(display_error_message(QMediaPlayer::Error)));
	delete this->qt_media_player;
	this->qt_media_player = NULL;
	this->is_wait_for_open = false;
	this->mutex_qmedia.unlock();
}

void DispQtVideoPlayer::video_request_open(void *handle_ptr, char *utf8_filename)
{
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	this->is_wait_for_open = true;
	this->ptr_handle_new = handle_ptr;
	this->media_name = QString::fromUtf8(utf8_filename);
	this->mutex_qmedia.unlock();
	emit this->signal_video_open();
}

void DispQtVideoPlayer::video_open(void)
{
	if(!this->video_widget)
		return;
	this->mutex_player.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	this->video_stop(this->ptr_handle_curr, false);
	this->video_mediaplayer_open();
	this->video_widget->setUpdatesEnabled(false);
	try {
		this->media_url = QUrl::fromLocalFile(this->media_name);
		this->media_content = QMediaContent(this->media_url);
		if(!this->media_content.isNull()) {
			this->qt_media_player->setMedia(this->media_content);
			if(this->qt_media_player->error() == QMediaPlayer::NoError)
				this->ptr_handle_curr = this->ptr_handle_new;
			else
				this->video_mediaplayer_close(); // FIXME: ???
		}
	} catch(int e) {
		mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "EXCEPTION video_open %d", e);
	}
	mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "OPEN err:%d %8.8X \"%s\"", ((this->qt_media_player)? (int)this->qt_media_player->error() : -99), (unsigned long)this->ptr_handle_curr, this->media_name.toUtf8().data());
	this->mutex_player.unlock();
}

bool DispQtVideoPlayer::video_wait_for_open(void)
{
	int counter = DISPQT_VIDEO_OPEN_TIMEOUT / 10; // ca. 5 secs
	this->main_window->mainwin_statusbar_msg_send((char *)"Waiting for video open...");
	while(this->is_wait_for_open && (--counter) && (pds_look_extgetch()!=KEY_ESC)) {pds_threads_sleep(10);}
	this->main_window->mainwin_statusbar_msg_send((this->is_video_open)? (char *)"OK" : (char *)"Video open failed!");
	return this->is_video_open;
}

void DispQtVideoPlayer::video_request_close(void *handle_ptr)
{
	this->mutex_player.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	this->ptr_handle_close = handle_ptr;
	this->is_wait_for_close = true;
	this->mutex_player.unlock();
	emit this->signal_video_close();
}

void DispQtVideoPlayer::video_close(void)
{
	this->mutex_player.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(!this->ptr_handle_close){
		this->mutex_player.unlock();
		return;
	}
	this->is_video_closed = false;
	this->video_stop(this->ptr_handle_close, true);
	this->ptr_handle_close = NULL;
	this->is_video_closed = !this->is_video_open;
	this->is_wait_for_close = false;
	this->mutex_player.unlock();
}

bool DispQtVideoPlayer::video_wait_for_close(void)
{
	int counter = DISPQT_VIDEO_OPEN_TIMEOUT / 10; // ca. 5 secs
	while(this->is_wait_for_close && (--counter) && (pds_look_extgetch()!=KEY_ESC)) {pds_threads_sleep(10);}
	return this->is_video_closed;
}

void DispQtVideoPlayer::video_stop(void *handle_ptr, bool close_media)
{
	mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "CLOSE co:%d h:%8.8X c:%8.8X", (int)close_media, (unsigned long)handle_ptr, (unsigned long)this->ptr_handle_curr);
	if(handle_ptr && handle_ptr != this->ptr_handle_curr)
		return;
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	this->is_playing = false;
	this->ptr_handle_curr = NULL;
	if(this->is_video_open) {
		this->is_video_open = false;
		if(this->qt_media_player)
			this->qt_media_player->stop();
	}
	this->mutex_qmedia.unlock();
	if(close_media)
		this->video_mediaplayer_close();
}

void DispQtVideoPlayer::status_changed(QMediaPlayer::MediaStatus status)
{
	this->video_cursor_select(status);

    switch (status) {
    	case QMediaPlayer::UnknownMediaStatus:
    	case QMediaPlayer::NoMedia:
    		this->is_video_open = false;
    		break;
    	case QMediaPlayer::LoadingMedia:
    		this->main_window->mainwin_statusbar_msg_send((char *)"Loading...");
    		break;
    	case QMediaPlayer::LoadedMedia:
    	case QMediaPlayer::BufferedMedia:
    		if(this->is_wait_for_open) {
    			this->video_widget->setUpdatesEnabled(true);
    			this->is_video_open = true;
    			this->is_wait_for_open = false;
    		}
    		break;
    	case QMediaPlayer::BufferingMedia:
    		this->main_window->mainwin_statusbar_msg_send((char *)"Buffering...");
    		break;
    	case QMediaPlayer::StalledMedia:
    		this->main_window->mainwin_statusbar_msg_send((char *)"Media Stalled");
    		this->is_video_open = false;
    		break;
    	case QMediaPlayer::EndOfMedia:
    		this->is_video_open = false;
    		mvps.adone = ADONE_EOF;  // !!!
    		break;
    	case QMediaPlayer::InvalidMedia:
    		display_error_message(QMediaPlayer::FormatError);
    		break;
    }
    mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "STATUS: %d", (int)status);
}

void DispQtVideoPlayer::video_cursor_select(QMediaPlayer::MediaStatus status)
{
#ifndef QT_NO_CURSOR
    if (status == QMediaPlayer::LoadingMedia ||
        status == QMediaPlayer::BufferingMedia ||
        status == QMediaPlayer::StalledMedia)
        setCursor(QCursor(Qt::BusyCursor));
    else
        unsetCursor();
#endif
}

void DispQtVideoPlayer::display_error_message(QMediaPlayer::Error err)
{
	if(this->qt_media_player){
		this->main_window->mainwin_statusbar_msg_send((char *)this->qt_media_player->errorString().toUtf8().data());
		mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "ERROR: %d \"%s\"", (int)err, (char *)this->qt_media_player->errorString().toUtf8().data());
	}
	if(err != QMediaPlayer::NoError) {
		if(this->is_video_open)
			mvps.adone = ADONE_EOF;  // !!! ???
		this->video_mediaplayer_close();
	}
}

void DispQtVideoPlayer::video_play(bool startflag)
{
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(this->is_video_open && this->qt_media_player){
		this->qt_media_player->play();
		this->is_playing = true;
	}
	this->mutex_qmedia.unlock();
	struct mmc_dispqt_config_s *gcfg = this->main_window->gui_config;
	unsigned long video_control = gcfg->video_control;
	if(this->is_playing && !(gcfg->transparent_control & DISPQT_CONFIG_TRANSCTRL_DESKTOPTRANS_ENABLED) && ((startflag && (video_control & DISPQT_CONFIG_VIDEOCTRL_PLAYSTART_FULLSCREEN)) || (video_control & DISPQT_CONFIG_VIDEOCTRL_ALWAYS_FULLSCREEN)))
		emit this->signal_video_set_fullscreen(true);
}

void DispQtVideoPlayer::video_pause(void)
{
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(this->is_video_open && this->qt_media_player)
		this->qt_media_player->pause();
	this->is_playing = false;
	this->mutex_qmedia.unlock();
}

void DispQtVideoPlayer::video_seek_ms(qint64 pos_ms)
{
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(this->is_video_open && this->qt_media_player)
		this->qt_media_player->setPosition(pos_ms);
	this->mutex_qmedia.unlock();
	mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "SEEK: %d", (int)pos_ms);
}

qint64 DispQtVideoPlayer::video_tell_ms(void)
{
	qint64 pos_ms;
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	pos_ms = ((this->is_video_open && this->qt_media_player)? this->qt_media_player->position() : -1);
	this->mutex_qmedia.unlock();
	//mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "TELL: %d", (int)pos_ms);
	return pos_ms;
}

qint64 DispQtVideoPlayer::video_duration_ms(void)
{
	qint64 dur_ms;
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	dur_ms = ((this->is_video_open && this->qt_media_player)? this->qt_media_player->duration() : -1);
	this->mutex_qmedia.unlock();
	mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "DUR: %d", (int)dur_ms);
	return dur_ms;
}

void DispQtVideoPlayer::video_set_speed(int speed) // 1000 = 100%
{
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(this->is_video_open && this->qt_media_player){
		if(speed < 100)
			speed = 100;
		else if(speed > 10000)
			speed = 10000;
		qreal rate = (qreal)speed / 1000.0;
		this->qt_media_player->setPlaybackRate(rate);
		mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "SPD: %d %1.3f", speed, rate);
	}
	this->mutex_qmedia.unlock();
}

void DispQtVideoPlayer::audio_set_volume(int vol)
{
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(this->is_video_open && this->qt_media_player) {
		if(vol < 0)
			vol = 0;
		else if(vol > 100)
			vol = 100;
		this->qt_media_player->setVolume(vol);
		mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "VOL: %d", vol);
	}
	this->mutex_qmedia.unlock();
}

void DispQtVideoPlayer::video_set_fullscreen(bool set_fs)
{
	if(!this->video_widget)
		return;
	bool is_fs = this->video_widget->isFullScreen();
	if(set_fs != is_fs)
		this->video_switch_fullscreen();
}

void DispQtVideoPlayer::video_switch_fullscreen(void)
{
	if(!this->video_widget)
		return;
	bool full = !this->video_widget->isFullScreen();
	if(full && !this->qt_media_player)
		return;
	if(full)
	    emit this->main_window->signal_video_fullscreen(full);
	else
	    emit this->signal_video_apply_fullscreen(full);
}

void DispQtVideoPlayer::video_reset_widget(void)  // FIXME: this causes play gap
{
	DispQtVideoWidget *dqvw_old = this->video_widget, *dqvw_new;
	if(!dqvw_old)
		goto err_out_reset;
	disconnect(dqvw_old->video_surface_contextmenu, SIGNAL(aboutToHide()), this, SLOT(cmenu_about_hide()));
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	dqvw_new = new DispQtVideoWidget(this->main_window, this);
	if(!dqvw_new)
		goto err_out_reset;
	this->video_layout->setEnabled(false);
	this->video_layout->removeWidget(dqvw_old);
	this->video_widget = dqvw_new;
	if(this->qt_media_player)
		this->qt_media_player->setVideoOutput(dqvw_new);
	this->video_layout->addWidget(dqvw_new);
	this->main_window->mainwin_vmixer_reconfig();
	dqvw_new->mouse_cursor_forbid_hide(dqvw_old->get_mousecursor_forbidden_hide());
	this->video_layout->setEnabled(true);
	delete dqvw_old;
err_out_reset:
	this->mutex_qmedia.unlock();
}

void DispQtVideoPlayer::video_apply_fullscreen(bool full)
{

	struct mmc_dispqt_config_s *gcfg = this->main_window->gui_config;
	if(!(gcfg->mainwin_control & DISPQT_CONFIG_MAINWINCTRL_SHOW_VIDEO)){
		this->video_widget->widget_control((full)? DISPQT_CONFIG_MAINWINCTRL_SHOW_VIDEO : gcfg->mainwin_control);
		if(full && ((gcfg->video_control & (DISPQT_CONFIG_VIDEOCTRL_SHOW_MAINWIN_ON_VIDEO|DISPQT_CONFIG_VIDEOCTRL_FULLSCREEN_VIDEO_RESET)) == (DISPQT_CONFIG_VIDEOCTRL_SHOW_MAINWIN_ON_VIDEO|DISPQT_CONFIG_VIDEOCTRL_FULLSCREEN_VIDEO_RESET))
		){
			this->video_reset_widget();
		}
	}
	this->video_widget->setFullScreen(full);
	if(!full && (gcfg->mainwin_control & DISPQT_CONFIG_MAINWINCTRL_SHOW_VIDEO)
	 && ((gcfg->video_control & (DISPQT_CONFIG_VIDEOCTRL_SHOW_MAINWIN_ON_VIDEO|DISPQT_CONFIG_VIDEOCTRL_FULLSCREEN_VIDEO_RESET)) == (DISPQT_CONFIG_VIDEOCTRL_SHOW_MAINWIN_ON_VIDEO|DISPQT_CONFIG_VIDEOCTRL_FULLSCREEN_VIDEO_RESET))
	){
		this->video_reset_widget();
	}
	emit this->signal_video_apply_windows();
}

void DispQtVideoPlayer::video_check_windows(void)
{
    if(!this->video_widget)
        return;
    if(this->video_widget->isFullScreen() && this->qt_media_player)
        this->main_window->mainwin_video_check_show(QCursor::pos().x(), QCursor::pos().y());
    else
        this->video_widget->mouse_cursor_show_timed(false);
}

void DispQtVideoPlayer::video_apply_windows(void)
{
	if(!this->video_widget)
		return;
	this->video_check_windows();
	if(!this->video_widget->isFullScreen() || !this->qt_media_player) {
	    this->main_window->mainwin_video_check_show(-1, -1); // FIXME: better idea? (to hide unnecessary window(s) at non-fullscreen)
		emit this->main_window->signal_video_fullscreen(false);
	}
}

void DispQtVideoPlayer::video_handle_event_stop(void)
{
	if(this->main_window->gui_config->video_control & DISPQT_CONFIG_VIDEOCTRL_EXIT_FULLSCREEN_AT_STOP)
		this->video_set_fullscreen(false);
	else
		emit this->main_window->signal_video_mainwin_wakeup(true, false);
}

void DispQtVideoPlayer::video_mouse_cursor_show(bool timed)
{
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(this->video_widget)
		this->video_widget->mouse_cursor_show_timed(timed);
	this->mutex_qmedia.unlock();
}

void DispQtVideoPlayer::video_mouse_cursor_forbid_hide(bool forbid)
{
	this->mutex_qmedia.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(this->video_widget)
		this->video_widget->mouse_cursor_forbid_hide(forbid);
	this->mutex_qmedia.unlock();
}

//----------------------------------------------------------------------------

DispQtVideoWidget::DispQtVideoWidget(MainWindow *main_window, QWidget *parent) : QVideoWidget(parent)
{
	this->main_window = main_window;
	this->parent_widget = parent;
	this->setFocusPolicy(Qt::ClickFocus);

	setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Ignored);
	this->resize(150, 100);

	QPalette p = palette();
	p.setColor(QPalette::Window, Qt::black);
	setPalette(p);

	setAttribute(Qt::WA_OpaquePaintEvent);

	timer_singleclick = new QTimer(this);
	timer_singleclick->setInterval(DISPQT_VIDEOWIDGET_SINGLECLICK_TIMEOUTMS);
	timer_singleclick->setSingleShot(true);
	connect(timer_singleclick, SIGNAL(timeout()), this, SLOT(mouse_singleclick_timeouthandler()));
	this->timer_doubleclick.setSingleShot(true);
	this->singleclick_event = this->mouse_cursor_forbidden_hide = this->singleclick_fullsrc_start = this->fullsrc_curr = this->is_cursor_blank = false;

	this->timer_mouse_hide = new QTimer;
	this->timer_mouse_hide->setSingleShot(true);
	connect(this->timer_mouse_hide, SIGNAL(timeout()), this, SLOT(mouse_cursor_hide()));

	this->video_surface_contextmenu = new QMenu("Video Context Menu", this);
	this->video_surface_contextmenu->setStyleSheet("QMenu { font-size:13px; border: 1px solid transparent;}");

	//this->menuaction_filename = this->video_surface_contextmenu->addAction("");
	//this->menuaction_filename->setFont(QFont("", -1, QFont::DemiBold, false));
	//this->video_surface_contextmenu->addSeparator();

	action_play_start = this->video_surface_contextmenu->addAction(QIcon(":/images/player_play.png"), tr("&Play / Pause"));
	action_play_start->setShortcut(QKeySequence(tr("P")));
	action_play_stop = this->video_surface_contextmenu->addAction(QIcon(":/images/player_stop.png"), tr("&Stop playing"));
	action_play_stop->setShortcut(QKeySequence(tr("S")));
	action_play_skipback = this->video_surface_contextmenu->addAction(QIcon(":/images/player_skip_back.png"), tr("&Skip back"));
	action_play_skipback->setShortcut(QKeySequence(tr("-")));
	action_play_skipforward = this->video_surface_contextmenu->addAction(QIcon(":/images/player_skip_forward.png"), tr("&Skip forward"));
	action_play_skipforward->setShortcut(QKeySequence(tr("+")));
	if((this->main_window->gui_config->video_control & (DISPQT_CONFIG_VIDEOCTRL_SHOW_MAINWIN_ON_VIDEO|DISPQT_CONFIG_VIDEOCTRL_FULLSCREEN_VIDEO_RESET)) != (DISPQT_CONFIG_VIDEOCTRL_SHOW_MAINWIN_ON_VIDEO|DISPQT_CONFIG_VIDEOCTRL_FULLSCREEN_VIDEO_RESET))
	{
		action_view_videofs = this->video_surface_contextmenu->addAction(QIcon(":/images/monitor_green.png"), tr("&Fullscreen switch"));
		action_view_videofs->setShortcut(QKeySequence(tr("Alt+Enter")));
		connect(action_view_videofs, SIGNAL(triggered()), main_window, SLOT(actfunc_options_videofs()));
	}
	this->video_surface_contextmenu->addSeparator();
	action_file_exit = this->video_surface_contextmenu->addAction(tr("E&xit"));

	connect(action_play_start, SIGNAL(triggered()), this, SLOT(play_start_or_pause()));
	connect(action_play_stop, SIGNAL(triggered()), main_window, SLOT(play_stop()));
	connect(action_play_skipback, SIGNAL(triggered()), main_window, SLOT(play_skip_back()));
	connect(action_play_skipforward, SIGNAL(triggered()), main_window, SLOT(play_skip_forward()));
	connect(action_file_exit, SIGNAL(triggered()), main_window, SLOT(CloseWindowInternal()));

	connect(this->video_surface_contextmenu, SIGNAL(aboutToHide()), this, SLOT(cmenu_about_hide()));

	this->timer_cmenu_hide.setSingleShot(true);
}

void DispQtVideoWidget::widget_control(unsigned long mainwin_control)
{
	mutex_video.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	if(mainwin_control & DISPQT_CONFIG_MAINWINCTRL_SHOW_VIDEO){
		setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Ignored);
		this->resize(150, 100);
		this->show();
	} else
		this->hide();
	mutex_video.unlock();
}

void DispQtVideoWidget::mouse_cursor_forbid_hide(bool forbid)
{
	this->mouse_cursor_forbidden_hide = forbid;
	this->mouse_cursor_show_timed(false);
}

void DispQtVideoWidget::mouse_cursor_show_timed(bool timed)
{
	mutex_video.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	mpxplay_debugf(DISPQT_DEBUGOUT_CURSOR, "mouse_cursor_show_timed b:%d t:%d", this->is_cursor_blank, timed);
	if(this->is_cursor_blank){
		QApplication::restoreOverrideCursor();
		this->is_cursor_blank = false;
	}
	if(timed && !this->mouse_cursor_forbidden_hide && this->isFullScreen()){
		this->timer_mouse_hide->setInterval(DISPQT_VIDEO_CURSORHIDE_TIMEOUT);
		this->timer_mouse_hide->start();
	}else
		this->timer_mouse_hide->stop();
	this->setMouseTracking(this->isFullScreen());
	mutex_video.unlock();
}

void DispQtVideoWidget::mouse_cursor_hide(void)
{
	mutex_video.tryLock(DISPQT_VIDEO_MUTEX_TIMEOUT);
	mpxplay_debugf(DISPQT_DEBUGOUT_CURSOR, "mouse_cursor_hide b:%d", this->is_cursor_blank);
	if(!this->is_cursor_blank){
		QApplication::setOverrideCursor(Qt::BlankCursor);
		this->is_cursor_blank = true;
	}
	mutex_video.unlock();
}

void DispQtVideoWidget::play_start_or_pause(void)
{
	pds_pushkey(kb[50].c); // KeyPlayPause1
}

void DispQtVideoWidget::keyPressEvent(QKeyEvent *event)
{
    mpxplay_debugf(stdout, "KEY event DispQtVideoWidget");
    this->main_window->handle_keypressevent(event);
}

void DispQtVideoWidget::mouseDoubleClickEvent(QMouseEvent *event)
{
	this->timer_doubleclick.start(DISPQT_VIDEOWIDGET_DOUBLECLICK_TIMEOUTMS);
	this->timer_singleclick->stop();
	this->fullsrc_curr = !isFullScreen();
	emit ((DispQtVideoPlayer *)this->parent_widget)->signal_video_switch_fullscreen();
	if(this->singleclick_event) {
		this->main_window->mainwin_handle_keypresscode(kb[51].c);
		this->singleclick_event = false;
	}
	//mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "DC se:%d", (int)this->singleclick_event);
	event->accept();
}

void DispQtVideoWidget::mouse_singleclick_timeouthandler(void)
{
	if(!this->timer_doubleclick.isActive() && (this->singleclick_fullsrc_start == this->fullsrc_curr)) {
		this->timer_doubleclick.start(DISPQT_VIDEOWIDGET_DOUBLECLICK_TIMEOUTMS);
		this->main_window->mainwin_handle_keypresscode(kb[51].c);
		//mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "STO");
		this->singleclick_event = true;
	}
}

void DispQtVideoWidget::mousePressEvent(QMouseEvent *event)
{
	if(event->button() & Qt::LeftButton){  // start or stop playing with single click, fullscreen switch with double click
	    if(!this->timer_doubleclick.isActive() && !this->timer_singleclick->isActive() && !this->timer_cmenu_hide.isActive()) {
	        this->timer_singleclick->start(DISPQT_VIDEOWIDGET_SINGLECLICK_TIMEOUTMS);
	        this->singleclick_fullsrc_start = this->fullsrc_curr = isFullScreen();
	        this->singleclick_event = false;
	        //mpxplay_debugf(DISPQT_DEBUG_OUTPUT, "MPE");
	    }
	} else if(event->button() & Qt::RightButton){
		//this->menuaction_filename->setText(((DispQtVideoPlayer *)this->parent_widget)->video_media_name());
		this->video_surface_contextmenu->popup(event->globalPos());
		this->mouse_cursor_show_timed(false);
	    //this->main_window->handle_mousepressevent(event);
	}
	event->accept();
}

void DispQtVideoWidget::wheelEvent(QWheelEvent* event)
{
	mpxplay_debugf(DISPQT_DEBUGOUT_EVENT, "video wheel event %d", event->delta());
	this->main_window->mainwin_handle_keypresscode(kb[(event->delta() < 0)? 83 : 81].c);  // volume up / down
	if(this->main_window->gui_config->mainwin_control & DISPQT_CONFIG_MAINWINCTRL_SHOW_AVMIXER)
		emit this->main_window->signal_video_mainwin_wakeup(true, true);
	QVideoWidget::wheelEvent(event);
}

void DispQtVideoWidget::mouseMoveEvent(QMouseEvent *event)
{
	if(isFullScreen())
		this->main_window->mainwin_video_check_show(event->globalX(), event->globalY());
	QVideoWidget::mouseMoveEvent(event);
}

bool DispQtVideoWidget::event(QEvent *event)
{
	//mpxplay_debugf(DISPQT_DEBUGOUT_EVENT, "video event: %d", event->type());
	switch(event->type()){
		case QEvent::Close:
			if(this->main_window->gui_config->video_control & DISPQT_CONFIG_VIDEOCTRL_ENABLE_VIDEO)
				this->main_window->CloseWindowTriggeredfromExternal();
			break;
	}
	return QVideoWidget::event(event);
}
